#!/usr/bin/env python3
import json
import logging
import ntpath
from datetime import datetime
from typing import Iterable, Optional, Union
from uuid import UUID

from google.cloud import storage  # type: ignore

from .constant import BUCKETS, PATH_TEMPLATES, FileType
from .env import get_blob_env
from .utils import file_exist, iter_files

logger = logging.getLogger(__name__)


def _client(credentials: Union[str, dict]) -> storage.Client:
    """supports dict, string or path to the JSON file"""
    if isinstance(credentials, dict):
        return storage.Client.from_service_account_info(credentials)
    if file_exist(credentials):
        return storage.Client.from_service_account_json(credentials)
    if isinstance(credentials, str):
        credentials = json.loads(credentials)
        return storage.Client.from_service_account_info(credentials)
    raise ValueError("needs path or dict for credentials")


def _path(source_id: UUID, file_type: FileType, file_path: str) -> str:
    now = datetime.utcnow()
    timestamp = int(now.timestamp())
    filename = ntpath.basename(file_path)

    path_template = PATH_TEMPLATES[file_type]
    return path_template.format(
        timestamp=timestamp,
        source_id=source_id,
        filename=filename,
    )


def _get_blob(
    credentials: Union[str, dict],
    source_id: UUID,
    file_path: str,
    file_type: FileType,
) -> storage.Blob:
    """get the target blob to upload to"""
    client = _client(credentials)
    bucket_name = BUCKETS[file_type]
    path = _path(source_id, file_type, file_path)

    bucket = client.bucket(bucket_name)
    return bucket.blob(path)


def _upload(
    credentials: Union[str, dict],
    source_id: UUID,
    file_path: str,
    file_type: FileType,
) -> None:
    """
    credentials: path to file or dict
    source_id: id for the source
    file_type: type of file to upload
    file_path: path to the local file to upload
    """
    timeout, retries = get_blob_env()

    blob = _get_blob(credentials, source_id, file_path, file_type)
    with open(file_path, "rb") as f:
        blob.upload_from_file(f, timeout=timeout, num_retries=retries)
    logger.info(
        f"uploaded {file_path} as {file_type.value} to {blob.public_url}"
    )


def upload_manifest(
    credentials: Union[str, dict], source_id: UUID, file_path: str
) -> None:
    """
    credentials: path to file or dict
    source_id: id for the source
    file_path: path to the local manifest to upload
    """
    _upload(credentials, source_id, file_path, FileType.DBT)


def upload(
    credentials: Union[str, dict],
    source_id: UUID,
    file_type: FileType,
    file_path: Optional[str] = None,
    directory_path: Optional[str] = None,
) -> None:
    """
    credentials: path to file or dict
    source_id: id for the source
    file_type: type of file(s) uploaded - see FileType Enum
    file_path: path to the local visualization or warehouse file to upload
    directory_path: path to the local directory containing files to upload
    """
    files: Iterable[str]
    if directory_path:
        files = iter_files(directory_path)
    elif file_path:
        files = [file_path]
    else:
        message = "either file_path or directory_path should be defined"
        raise ValueError(message)

    for file_ in files:
        _upload(credentials, source_id, file_, file_type)



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import mode  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-H", "--host", help="Mode Analytics host")
    parser.add_argument("-w", "--workspace", help="Mode Analytics workspace")
    parser.add_argument(
        "-t", "--token", help="The Token value from the API token"
    )
    parser.add_argument(
        "-s", "--secret", help="The Password value from the API token"
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()

    client = mode.Client(
        host=args.host,
        workspace=args.workspace,
        token=args.token,
        secret=args.secret,
    )

    mode.extract_all(
        client,
        output_directory=args.output,
    )



from enum import Enum


class LookerAsset(Enum):
    """Looker assets"""

    DASHBOARDS = "dashboards"
    EXPLORES = "explores"
    FOLDERS = "folders"
    LOOKS = "looks"
    LOOKML_MODELS = "lookml_models"
    USERS = "users"
    CONNECTIONS = "connections"
    PROJECTS = "projects"



import logging
from typing import Optional

from .constants import DEFAULT_HANDLE, JsonRpcMethod
from .error import raise_for_error
from .websocket import WebsocketConnection

logger = logging.getLogger(__name__)


class JsonRpcClient:
    """
    JSON-RPC client connecting to QlikSense through websocket

    doc: https://qlik.dev/apis/json-rpc/qix
    """

    VERSION = "2.0"

    def __init__(self, websocket: WebsocketConnection):
        self.websocket = websocket
        self.call_id = 0

    def _increment_call_id(self):
        self.call_id += 1

    def _format_message(
        self, method: JsonRpcMethod, handle: int, params: Optional[list] = None
    ) -> dict:
        params = params or list()
        message = {
            "jsonrpc": self.VERSION,
            "method": method.value,
            "id": self.call_id,
            "handle": handle,
            "params": params,
        }
        return message

    def send_message(
        self,
        method: JsonRpcMethod,
        handle: int = DEFAULT_HANDLE,
        params: Optional[list] = None,
    ) -> dict:
        """Sends JSON-RPC message through websocket and checks no error"""
        self._increment_call_id()
        message = self._format_message(method, handle, params)
        response = self.websocket.call(message=message)
        raise_for_error(message, response)
        return response



import datetime
from typing import Optional

from ...utils import current_date


class TimeFilter:
    """
    Holds day and optional hours range to manage time filtering in ExtractionQueries
    Provides a default time filter (previous day)
    """

    def __init__(
        self,
        day: datetime.date,
        hour_min: Optional[int] = None,
        hour_max: Optional[int] = None,
    ):
        self.day = day
        self.hour_min = hour_min
        self.hour_max = hour_max

    @classmethod
    def default(cls):
        """
        Default time filter for extraction:
        - day = yesterday
        - hours = [0,23]
        """
        yesterday = current_date() - datetime.timedelta(days=1)
        return cls(day=yesterday, hour_min=0, hour_max=23)

    def to_dict(self) -> dict:
        return {
            "day": self.day,
            "hour_min": self.hour_min,
            "hour_max": self.hour_max,
        }



"""
Request timeout in seconds for Looker API
"""
DEFAULT_LOOKER_TIMEOUT_SECOND = 120
CASTOR_LOOKER_TIMEOUT_SECOND = "CASTOR_LOOKER_TIMEOUT_SECOND"

"""
Number of items per page when requesting Looker API
"""
DEFAULT_LOOKER_PAGE_SIZE = 500
CASTOR_LOOKER_PAGE_SIZE = "CASTOR_LOOKER_PAGE_SIZE"

# env variables
BASE_URL = "CASTOR_LOOKER_BASE_URL"
CLIENT_ID = "CASTOR_LOOKER_CLIENT_ID"
CLIENT_SECRET = "CASTOR_LOOKER_CLIENT_SECRET"
ALL_LOOKS = "CASTOR_LOOKER_ALL_LOOKS"



from unittest.mock import patch

from .constants import JsonRpcMethod
from .json_rpc import JsonRpcClient
from .websocket import WebsocketConnection


def test_json_rpc_client__send_message():

    dummy_server_url = "toto.fr"
    dummy_api_key = "IAmNotATrueSecretKey"
    ws = WebsocketConnection(server_url=dummy_server_url, api_key=dummy_api_key)

    client = JsonRpcClient(websocket=ws)
    any_method = JsonRpcMethod.OPEN_DOC

    mock_response = {"data": "all qlik"}
    with patch.object(WebsocketConnection, "call") as mock_call:
        mock_call.return_value = mock_response

        assert client.call_id == 0

        response = client.send_message(any_method)
        assert response == mock_response
        assert client.call_id == 1
        expected = {
            "jsonrpc": "2.0",
            "method": "OpenDoc",
            "id": 1,
            "handle": -1,
            "params": [],
        }
        mock_call.assert_called_with(message=expected)

        client.send_message(any_method, handle=12, params=["toto"])
        assert client.call_id == 2
        expected = {
            "jsonrpc": "2.0",
            "method": "OpenDoc",
            "id": 2,
            "handle": 12,
            "params": ["toto"],
        }
        mock_call.assert_called_with(message=expected)



from unittest.mock import patch

from .files import explode, search_files


def test_explode():
    checks = {
        "/some/path/file.csv": ("/some/path", "file", "csv"),
        "/file.json": ("/", "file", "json"),
        "/without/extension/file": ("/without/extension", "file", ""),
        "file.txt": ("", "file", "txt"),
        "/tmp/.bashrc": ("/tmp", ".bashrc", ""),
    }
    for path, expected in checks.items():
        assert explode(path) == expected


def test_search_files():
    file_list = (
        "foo.csv",
        "bar.csv",
        "unknown.csv",
        "123-foo.csv",
        "foo.txt",
        "bar.json",
        "deprecated-foo.csv",
    )

    with patch("glob.glob") as mocked:
        mocked.return_value = file_list
        directory = "/tmp"

        # no filters
        files = search_files(directory)
        assert set(files) == set(file_list)

        # filter extensions
        files = search_files(
            directory,
            filter_extensions={"json", "txt"},
        )
        assert set(files) == {"bar.json", "foo.txt"}

        # endswith + does not contain
        files = search_files(
            directory,
            filter_endswith="foo",
            does_not_contain={"deprecated"},
        )
        assert set(files) == {"foo.csv", "foo.txt", "123-foo.csv"}

        # no results
        files = search_files(
            directory,
            filter_extensions={"json"},
            does_not_contain={"bar"},
        )
        assert not files



from .asset import (
    CATALOG_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    SupportedAssets,
    WarehouseAsset,
    WarehouseAssetGroup,
)
from .client import AbstractWarehouseClient, SqlalchemyClient
from .extract import ExtractionProcessor, common_args
from .query import (
    QUERIES_DIR,
    AbstractQueryBuilder,
    ExtractionQuery,
    TimeFilter,
)



from typing import Tuple

import pytest

from .string import string_to_tuple


def _test(symbols: str, input_: str) -> Tuple[str, ...]:
    return string_to_tuple(symbols[0] + input_ + symbols[1])


def test__string_to_tuple():
    """Test method string_to_tuple"""

    # loop on supported symbols surrounding elements
    for symbols in ("[]", "{}", "()", "  "):

        # empty list, set, tuple and empty string
        # [], {}, (), ''
        assert _test(symbols, "") == tuple()

        # single string, simple quotes
        assert _test(symbols, "'a'") == ("a",)
        # single string, double quotes
        assert _test(symbols, '"a"') == ("a",)
        # single string, no quotes
        assert _test(symbols, "a") == ("a",)
        # single strings, double quotes and coma inside
        assert _test(symbols, '"Hi, how are you?"') == ("Hi, how are you?",)

        # multiple strings, simple quotes
        assert _test(symbols, "'a', 'b'") == ("a", "b")
        # multiple strings, double quotes
        assert _test(symbols, '"a", "b"') == ("a", "b")
        # multiple strings, double quotes and coma inside
        assert _test(symbols, '"Hi, how are you?", "I am fine, thanks!"') == (
            "Hi, how are you?",
            "I am fine, thanks!",
        )

        # multiple strings + carriage returns
        value = """ [
            'a',
            'b'
        ]
        """
        assert _test(symbols, value) == ("a", "b")

        # multiple strings + extra coma
        assert _test(symbols, "'a', 'b',") == ("a", "b")

        # single integer
        assert _test(symbols, "1") == ("1",)
        # multiple integers
        assert _test(symbols, "1, 2, 3") == ("1", "2", "3")

        # single float
        assert _test(symbols, "3.14") == ("3.14",)
        # multiple floats
        assert _test(symbols, "3.14, -2.0") == ("3.14", "-2.0")

        # mixed data types
        assert _test(symbols, "'a', 12, -5.7") == ("a", "12", "-5.7")

        with pytest.raises(SyntaxError):
            # unbalanced quotes
            _test(symbols, '"a", b"')
        with pytest.raises(ValueError):
            # multiple strings, no quotes
            _test(symbols, "a, b")



from .constants import ACCESS_DENIED_ERROR_CODE, BASE_ERROR_MESSAGE


class JsonRpcError(Exception):
    """
    Error class to be raised when a JSON-RPC call response is an error
    """

    def __init__(self, message: dict, error: dict):
        self.message = message
        self.error = error
        error_message = BASE_ERROR_MESSAGE.format(message=message, error=error)
        super().__init__(error_message)

    @property
    def is_access_denied(self) -> bool:
        """
        Returns True when the error was due to Access Denied
        """
        return self.error["code"] == ACCESS_DENIED_ERROR_CODE


def raise_for_error(message: dict, response: dict) -> None:
    """Raises JsonRpcError when response is an error"""
    if "error" in response:
        raise JsonRpcError(message, response["error"])



import os
import unittest
from unittest import mock

import requests_mock  # type: ignore
import tableauserverclient as TSC  # type: ignore

from .....tableau import TableauAsset
from ....client import ApiClient, get_paginated_objects
from ....constants import TABLEAU_SERVER_VERSION
from ..utils import KEYS

TEST_ASSET_DIR = os.path.join(
    os.path.dirname(__file__), "../assets/graphql/metadata"
)

METADATA_1 = os.path.join(TEST_ASSET_DIR, "metadata_1_get.json")
METADATA_2 = os.path.join(TEST_ASSET_DIR, "metadata_2_get.json")


EXPECTED_PAGINATED = [
    {"id": "0039e5d5-25fa-196b-c66e-c0675839e0b0"},
    {"id": "00b191ce-6055-aff5-e275-c26610c8c4d6"},
]


class PaginatedObjectTests(unittest.TestCase):
    @mock.patch.dict(os.environ, KEYS)
    def setUp(self):
        self._client = ApiClient()
        self._client._server = TSC.Server("http://test")

        # Fake signin
        self._client._server._site_id = "dad65087-b08b-4603-af4e-2887b8aafc67"
        self._client._server._auth_token = "j80k54ll2lfMZ0tv97mlPvvSCRyD0DOM"
        self._client._server.version = TABLEAU_SERVER_VERSION

        self.baseurl = self._client._server.metadata.baseurl

    def test_paginated_object_with_datasources(self):
        with open(METADATA_1, "rb") as f:
            response_1_json = f.read().decode()

        with open(METADATA_2, "rb") as f:
            response_2_json = f.read().decode()

        with requests_mock.mock() as m:
            m.post(
                self.baseurl,
                [
                    {"text": response_1_json, "status_code": 200},
                    {"text": response_2_json, "status_code": 200},
                ],
            )

            results = get_paginated_objects(
                self._client._server,
                TableauAsset.DATASOURCE,
                self._client._page_size,
            )

        self.assertEqual(results, EXPECTED_PAGINATED)



import logging
from typing import Iterable, Optional, Tuple

from ...logger import add_logging_file_handler
from ...utils import (
    SafeMode,
    current_timestamp,
    deep_serialize,
    get_output_filename,
    write_json,
    write_summary,
)
from .api import ApiClient, Credentials, explore_names_associated_to_dashboards
from .assets import LookerAsset
from .parameters import get_parameters

logger = logging.getLogger(__name__)


def _safe_mode(directory: str) -> SafeMode:
    add_logging_file_handler(directory)
    return SafeMode((Exception,), float("inf"))


def _client(
    base_url: str,
    client_id: str,
    client_secret: str,
    timeout: Optional[int],
    safe_mode: Optional[SafeMode],
) -> ApiClient:
    credentials = Credentials(
        base_url=base_url,
        client_id=client_id,
        client_secret=client_secret,
        timeout=timeout,
    )
    return ApiClient(credentials=credentials, safe_mode=safe_mode)


def iterate_all_data(
    client: ApiClient, all_looks: bool
) -> Iterable[Tuple[LookerAsset, list]]:
    """Iterate over the extracted Data From looker"""

    logger.info("Extracting users from Looker API")
    yield LookerAsset.USERS, deep_serialize(client.users())

    logger.info("Extracting folders from Looker API")
    yield LookerAsset.FOLDERS, deep_serialize(client.folders())

    logger.info("Extracting looks from Looker API")
    yield LookerAsset.LOOKS, deep_serialize(client.looks(all_looks))

    logger.info("Extracting dashboards from Looker API")
    dashboards = client.dashboards()
    yield LookerAsset.DASHBOARDS, deep_serialize(dashboards)

    logger.info("Extracting lookml models from Looker API")
    lookmls = client.lookml_models()
    yield LookerAsset.LOOKML_MODELS, deep_serialize(lookmls)

    logger.info("Extracting explores from Looker API")
    explore_names = explore_names_associated_to_dashboards(lookmls, dashboards)
    yield LookerAsset.EXPLORES, deep_serialize(client.explores(explore_names))

    logger.info("Extracting connections from Looker API")
    connections = client.connections()
    yield LookerAsset.CONNECTIONS, deep_serialize(connections)

    logger.info("Extracting projects from Looker API")
    projects = client.projects()
    yield LookerAsset.PROJECTS, deep_serialize(projects)


def extract_all(**kwargs) -> None:
    """
    Extract Data From looker and store it locally in files under the
    output_directory
    """
    parameters = get_parameters(**kwargs)
    output_directory = parameters.output_directory
    base_url = parameters.base_url

    is_safe_mode = parameters.is_safe_mode
    safe_mode = _safe_mode(output_directory) if is_safe_mode else None
    client = _client(
        base_url=base_url,
        client_id=parameters.client_id,
        client_secret=parameters.client_secret,
        timeout=parameters.timeout,
        safe_mode=safe_mode,
    )

    ts = current_timestamp()

    for key, data in iterate_all_data(client, parameters.all_looks):
        filename = get_output_filename(key.value, output_directory, ts)
        write_json(filename, data)

    write_summary(output_directory, ts, base_url=base_url)



import pytest

from .constants import ACCESS_DENIED_ERROR_CODE
from .error import JsonRpcError, raise_for_error


def test_json_rpc_error__is_access_denied():
    any_message = {"hello": "world"}

    # no access denied
    another_code = ACCESS_DENIED_ERROR_CODE + 1
    error_msg = {"code": another_code}
    error = JsonRpcError(message=any_message, error=error_msg)
    assert not error.is_access_denied

    # access denied
    error_msg = {"code": ACCESS_DENIED_ERROR_CODE}
    error = JsonRpcError(message=any_message, error=error_msg)
    assert error.is_access_denied


def test_raise_for_error():
    any_message = {"hello": "world"}

    # no error
    response = {"everything is": "OK"}
    raise_for_error(any_message, response)

    # error
    response = {"error": "Houston, we have a problem"}
    with pytest.raises(JsonRpcError):
        raise_for_error(any_message, response)



import logging
import re
from enum import Enum
from typing import Optional

from sqlalchemy.engine import Connection

from ...utils import to_string_array, uri_encode
from ..abstract import SqlalchemyClient

SNOWFLAKE_URI = "snowflake://{user}:{password}@{account}/?application=castor"
PUBLIC_ROLE = "PUBLIC"

logger = logging.getLogger(__name__)


def _scalar(connection: Connection, query: str) -> Optional[str]:
    """returns the first column of the first row, None if no row"""
    scalar = connection.execute(query).scalar()
    return None if scalar is None else str(scalar)


class UseResource(Enum):
    """resources using the USE syntax"""

    ROLE = "ROLE"
    WAREHOUSE = "WAREHOUSE"


def _use(connection: Connection, resource: UseResource, name: str):
    """
    USE ROLE/WAREHOUSE do not accept SQL parameter substitution
        we do a sanity check prior to using it
    DON'T COPY THIS CODE FOR ANOTHER PURPOSE
    """
    target = resource.value
    if not re.fullmatch("[\w_-]+", name):
        raise NameError(f"{target} name invalid {name}")
    command = f'USE {target} "{name}"'
    logger.info(f"executing {command}")
    connection.execute(command)


def _first_non_public_role(connection: Connection) -> str:
    """fetch the first non public role granted to the user"""
    raw = _scalar(connection, "SELECT CURRENT_AVAILABLE_ROLES()")
    if not raw:
        raise ConnectionError("error at fetching granted roles")
    logger.info(f"available roles {raw}")
    parsed = to_string_array(raw)
    granted = {r.upper() for r in parsed} - {PUBLIC_ROLE}
    if not granted:
        raise ConnectionError("missing non public role")
    return next(iter(granted))


class SnowflakeClient(SqlalchemyClient):
    """snowflake client"""

    # hide information message from the connectors
    logging.getLogger("snowflake.connector").setLevel(logging.WARNING)

    def __init__(
        self,
        credentials: dict,
        warehouse: Optional[str] = None,
        role: Optional[str] = None,
    ):
        super().__init__(credentials)
        self._role = role
        self._warehouse = warehouse

    @staticmethod
    def name() -> str:
        return "Snowflake"

    def _engine_options(self, credentials: dict) -> dict:
        return {}

    def _build_uri(self, credentials: dict) -> str:
        return SNOWFLAKE_URI.format(
            user=credentials["user"],
            password=uri_encode(credentials["password"]),
            account=credentials["account"],
        )

    def role(self, connection: Connection) -> None:
        """check and set role"""
        if self._role:
            logger.info(f"Role was provided in arguments: {self._role}")
            _use(connection, UseResource.ROLE, self._role)
            return

        current = _scalar(connection, "SELECT CURRENT_ROLE()")
        if current and current.upper() != PUBLIC_ROLE:
            logger.info(f"Role is already set: {current}")
            return

        role = _first_non_public_role(connection)
        logger.info(f"Using first available role: {role}")
        _use(connection, UseResource.ROLE, role)

    def warehouse(self, connection: Connection):
        """check and set warehouse"""
        if self._warehouse:
            logger.info(
                f"Warehouse was provided in arguments: {self._warehouse}"
            )
            _use(connection, UseResource.WAREHOUSE, self._warehouse)
            return

        current = _scalar(connection, "SELECT CURRENT_WAREHOUSE()")
        if current:
            logger.info(f"Warehouse is already set: {current}")
            return

        warehouse = _scalar(connection, "SHOW WAREHOUSES")
        if not warehouse:
            raise ConnectionError("no warehouse available")
        logger.info(f"Using first available warehouse: {warehouse}")
        _use(connection, UseResource.WAREHOUSE, warehouse)

    def connect(self) -> Connection:
        """enhance default behaviour to check role and warehouse"""
        connection = super().connect()
        self.role(connection)
        self.warehouse(connection)
        return connection
